#region Using directives
using System;
using System.Globalization;
using System.IO;
using System.Reflection;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;
using System.Text;
using Microsoft.Win32;
using System.Security.Permissions;
using System.Drawing;
using System.Windows.Forms;
using System.Collections.Generic;
#endregion

namespace ShellExt
{
    #region ContextMenuItem
    /// <summary>
    /// Test event args
    /// </summary>
    /// <returns>Event args</returns>
    public class ExecuteItemEventArgs : EventArgs
    {
        public int index;
    } // class TestEventArgs


    /// <summary>
    /// Context menu item
    /// </summary>
    public class ContextMenuItem
    {
        public delegate void ExecuteItemEventHandler(object sender, ExecuteItemEventArgs ea);

        #region Properties
        private string text;

        public string Text
        {
            get { return text; }
            set { text = value; }
        }

        private string helpText;

        public string HelpText
        {
            get { return helpText; }
            set { helpText = value; }
        }

        private bool ownerDraw;

        public bool OwnerDraw
        {
            get { return ownerDraw; }
            set { ownerDraw = value; }
        }

        public Dictionary<string, string> Properties = new Dictionary<string, string>();

        #endregion

        #region Events
        private MeasureItemEventHandler measureItemEventHandler;

        public MeasureItemEventHandler OnMeasureItem
        {
            get { return measureItemEventHandler; }
            set { measureItemEventHandler = value; }
        }

        private DrawItemEventHandler drawItemEventHandler;

        public DrawItemEventHandler OnDrawItem
        {
            get { return drawItemEventHandler; }
            set { drawItemEventHandler = value; }
        }

        private ExecuteItemEventHandler executeItemEventHandler;

        public ExecuteItemEventHandler OnExecuteItem
        {
            get { return executeItemEventHandler; }
            set { executeItemEventHandler = value; }
        }
        #endregion

        #region Methods
        /// <summary>
        /// Do draw item
        /// </summary>
        /// <param name="ea">Ea</param>
        public void DoDrawItem(DrawItemEventArgs ea)
        {
            if (drawItemEventHandler != null)
                drawItemEventHandler(this, ea);
        } // DoDrawItem()

        /// <summary>
        /// Do measure item
        /// </summary>
        /// <param name="ea">Ea</param>
        public void DoMeasureItem(MeasureItemEventArgs ea)
        {
            if (measureItemEventHandler != null)
                measureItemEventHandler(this, ea);
        } // DoMeasureItem()

        /// <summary>
        /// Do execute item
        /// </summary>
        /// <param name="ea">Ea</param>
        public void DoExecuteItem(ExecuteItemEventArgs ea)
        {
            if (executeItemEventHandler != null)
                executeItemEventHandler(this, ea);
        } // DoExecuteItem()

        #endregion
    }
    #endregion

    #region Enumerations
    public enum WM
    {
        WM_DRAWITEM = 0x002B,
        WM_MEASUREITEM = 0x002C
    }
    #endregion

    #region Structures
    [StructLayout(LayoutKind.Sequential)]
    public struct DRAWITEMSTRUCT
    {
        public int CtlType;
        public int CtlID;
        public int itemID;
        public int itemAction;
        public int itemState;
        public IntPtr hwndItem;
        public IntPtr hDC;
        public RECT rcItem;
        public IntPtr itemData;
    }

    [StructLayout(LayoutKind.Sequential)]
    public struct RECT
    {
        public int left;
        public int top;
        public int right;
        public int bottom;
        public int Width
        {
            get { return right - left; }
        }

        public int Height
        {
            get { return bottom - top; }
        }

        public static implicit operator Rectangle(RECT r)
        {
            return new Rectangle(r.left, r.top, r.Width, r.Height);
        }
    }


    [StructLayout(LayoutKind.Sequential)]
    public struct MEASUREITEMSTRUCT
    {
        public int CtlType;
        public int CtlID;
        public int itemID;
        public int itemWidth;
        public int itemHeight;
        IntPtr itemData;
    }
    #endregion

    /// <summary>
    /// A base class for ContextMenuShell extensions that implement ownerdrawn functionality. This class is based on an article 
    /// by Dino Esposito, http://www.theserverside.net/articles/showarticle.tss?id=ShellExtensions
    /// </summary>
    public class ContextMenuShellExtension : IShellExtInit, IContextMenu, IContextMenu3
    {
        #region private methods
        /// <summary>
        /// Calculate id command
        /// </summary>
        /// <param name="idCommand">Id command</param>
        /// <returns>Int</returns>
        private int calculateIdCommand(int idCommand)
        {
            return (idCommand - usedIdCommandFirst);
        } // calculateIdCommand(idCommand)
        #endregion

        #region Protected Members
        protected IDataObject m_dataObject = null;
        protected uint m_hDrop = 0;
        protected List<ContextMenuItem> Items = new List<ContextMenuItem>();
        protected int usedIdCommandFirst;
        #endregion

        #region IShellExtInit
        /// <summary>
        /// Initializes the shell extension
        /// </summary>
        /// <param name="pidlFolder">Address of an ITEMIDLIST structure that uniquely identifies a folder. For property sheet extensions, this parameter is NULL. For shortcut menu extensions, it is the item identifier list for the folder that contains the item whose shortcut menu is being displayed. For nondefault drag-and-drop menu extensions, this parameter specifies the target folder.</param>
        /// <param name="lpdobj">Address of an IDataObject interface object that can be used to retrieve the objects being acted upon.</param>
        /// <param name="hKeyProgID">Registry key for the file object or folder type.</param>
        /// <returns>Returns NOERROR if successful, or an OLE-defined error value otherwise.</returns>
        int IShellExtInit.Initialize(IntPtr pidlFolder, IntPtr lpdobj, uint hKeyProgID)
        {
            try
            {
                m_dataObject = null;
                if (lpdobj != (IntPtr)0)
                {
                    // Get info about the directory
                    m_dataObject = (IDataObject)Marshal.GetObjectForIUnknown(lpdobj);
                    FORMATETC fmt = new FORMATETC();
                    fmt.cfFormat = CLIPFORMAT.CF_HDROP;
                    fmt.ptd = 0;
                    fmt.dwAspect = DVASPECT.DVASPECT_CONTENT;
                    fmt.lindex = -1;
                    fmt.tymed = TYMED.TYMED_HGLOBAL;
                    STGMEDIUM medium = new STGMEDIUM();
                    m_dataObject.GetData(ref fmt, ref medium);
                    m_hDrop = medium.hGlobal;

                    InitializeMenu();
                }
            }
            catch (Exception)
            {
            }
            return 0;
        }



        #endregion

        #region IContextMenu Members

        /// <summary>
        /// Queries the context menu.
        /// </summary>
        /// <param name="hMenu">Handle to the menu. The handler should specify this handle when adding menu items.</param>
        /// <param name="iMenu">Zero-based position at which to insert the first menu item.</param>
        /// <param name="idCmdFirst">Minimum value that the handler can specify for a menu item identifier.</param>
        /// <param name="idCmdLast">Maximum value that the handler can specify for a menu item identifier.</param>
        /// <param name="uFlags">Optional flags specifying how the shortcut menu can be changed.</param>
        /// <returns>
        /// If successful, returns an HRESULT value that has its severity value set to SEVERITY_SUCCESS and its code value set to the offset of the largest command identifier that was assigned, plus one. For example, assume that idCmdFirst is set to 5 and you add three items to the menu with command identifiers of 5, 7, and 8. The return value should be MAKE_HRESULT(SEVERITY_SUCCESS, 0, 8 - 5 + 1). Otherwise, it returns an OLE error value.
        /// </returns>
        public int QueryContextMenu(uint hMenu, uint iMenu, int idCmdFirst, int idCmdLast, uint uFlags)
        {
            int id = 1;
            usedIdCommandFirst = idCmdFirst;
            if ((uFlags & 0xf) == 0 || (uFlags & (uint)CMF.CMF_EXPLORE) != 0)
            {
                uint nselected = Helpers.DragQueryFile(m_hDrop, 0xffffffff, null, 0);
                // Multiple selection not supported in this version
                if (nselected == 1)
                {
                    StringBuilder sb = new StringBuilder(1024);
                    Helpers.DragQueryFile(m_hDrop, 0, sb, sb.Capacity + 1);
                    string filePath = sb.ToString();

                    if (AcceptContext(filePath))
                    {
                        MENUITEMINFO mii = new MENUITEMINFO();
                        mii.cbSize = 48;
                        mii.fMask = (uint)MIIM.ID | (uint)MIIM.TYPE | (uint)MIIM.STATE;

                        foreach (ContextMenuItem item in Items)
                        {
                            mii.wID = idCmdFirst + id;
                            // If user indicated ownerdraw set appropriate flag
                            if (item.OwnerDraw)
                                mii.fType = (uint)MF.OWNERDRAW;
                            else
                                mii.fType = (uint)MF.STRING;

                            mii.dwTypeData = item.Text;
                            mii.fState = (uint)MF.ENABLED;
                            Helpers.InsertMenuItem(hMenu, (uint)2, 1, ref mii);
                            id++;
                        }

                        // Only add seperator if there where items to add
                        if (Items.Count > 0)
                        {
                            mii.fType = (uint)MF.SEPARATOR;
                            mii.fState = (uint)MF.ENABLED;
                            Helpers.InsertMenuItem(hMenu, (uint)2, 1, ref mii);
                        }
                    }
                }
            }
            return id;
        }

        /// <summary>
        /// Invokes the command.
        /// </summary>
        /// <param name="pici">The pici.</param>
        public void InvokeCommand(IntPtr pici)
        {
            try
            {
                INVOKECOMMANDINFO ici = (INVOKECOMMANDINFO)Marshal.PtrToStructure(pici, typeof(ShellExt.INVOKECOMMANDINFO));
                if (ici.verb - 1 <= Items.Count)
                {
                    ContextMenuItem menuItem = Items[ici.verb - 1];
                    ExecuteItemEventArgs ea = new ExecuteItemEventArgs();
                    ea.index = ici.verb;
                    menuItem.DoExecuteItem(ea);
                } // if
            }
            catch (Exception)
            {
            }
        }

        /// <summary>
        /// Retrieves information about a shortcut menu command, including the help string and the language-independent, or canonical, name for the command.
        /// </summary>
        /// <param name="idCmd">Menu command identifier offset.</param>
        /// <param name="uFlags">Flags specifying the information to return.</param>
        /// <param name="reserved">Reserved. Applications must specify NULL when calling this method, and handlers must ignore this parameter when called.</param>
        /// <param name="commandString">Address of the buffer to receive the null-terminated string being retrieved.</param>
        /// <param name="cch">Size of the buffer, in characters, to receive the null-terminated string.</param>
        /// <returns>
        /// Returns NOERROR if successful, or an OLE-defined error value otherwise.
        /// </returns>
        public void GetCommandString(int idCmd, uint uFlags, int reserved, StringBuilder commandString, int cch)
        {
            switch (uFlags)
            {
                case (uint)GCS.VERB:
                    commandString = new StringBuilder(Items[idCmd - 1].Text.Substring(1, cch - 1));
                    break;
                case (uint)GCS.HELPTEXT:
                    commandString = new StringBuilder(Items[idCmd - 1].HelpText.Substring(1, cch));
                    break;
                case (uint)GCS.VALIDATE:
                    break;
            }

        }

        #endregion

        #region IContextMenu2 Members

        /// <summary>
        /// Allows client objects of the IContextMenu interface to handle messages associated with owner-drawn menu items.
        /// </summary>
        /// <param name="uMsg">Message to be processed. In the case of some messages, such as WM_INITMENUPOPUP, WM_DRAWITEM, WM_MENUCHAR, or WM_MEASUREITEM, the client object being called may provide owner-drawn menu items.</param>
        /// <param name="wParam">Additional message information. The value of this parameter depends on the value of the uMsg parameter.</param>
        /// <param name="lParam">Additional message information. The value of this parameter depends on the value of the uMsg parameter</param>
        /// <returns>Returns S_OK if successful, or an error value otherwise.</returns>
        public int HandleMenuMsg(uint uMsg, uint wParam, uint lParam)
        {
            // Just return 0, this message is never called as long as IContextMenu3 is implemented
            return 0;
        }

        #endregion

        #region IContextMenu3 Members
        /// <summary>
        /// Allows client objects of the IContextMenu3 interface to handle messages associated with owner-drawn menu items
        /// </summary>
        /// <param name="uMsg">Message to be processed. In the case of some messages, such as WM_INITMENUPOPUP, WM_DRAWITEM, WM_MENUCHAR, or WM_MEASUREITEM, the client object being called may provide owner-drawn menu items.</param>
        /// <param name="wParam">Additional message information. The value of this parameter depends on the value of the uMsg parameter.</param>
        /// <param name="lParam">Additional message information. The value of this parameter depends on the value of the uMsg parameter.</param>
        /// <param name="plResult">Address of an LRESULT value that the owner of the menu will return from the message. This parameter can be NULL.</param>
        /// <returns>Returns S_OK if successful, or an error value otherwise.</returns>
        public int HandleMenuMsg2(uint uMsg, IntPtr wParam, IntPtr lParam, IntPtr plResult)
        {
            Graphics gr;
            int cmdId;
            ContextMenuItem menuItem;
            switch (uMsg)
            {
                case (int)WM.WM_DRAWITEM:
                    // For menu's the wParam equals zero
                    if (wParam != IntPtr.Zero)
                        return 0;

                    DRAWITEMSTRUCT dis = (DRAWITEMSTRUCT)Marshal.PtrToStructure(lParam, typeof(DRAWITEMSTRUCT));

                    gr = Graphics.FromHdc(dis.hDC);
                    try
                    {
                        DrawItemState state = (DrawItemState)dis.itemState;

                        Color foreColor;
                        Color backColor;

                        // In case the menuitem is selected, use appropriate system colors as default
                        if ((state & DrawItemState.Selected) == DrawItemState.Selected)
                        {
                            backColor = SystemColors.Highlight;
                            foreColor = SystemColors.HighlightText;
                        }
                        else
                        {
                            backColor = SystemColors.Menu;
                            foreColor = SystemColors.MenuText;
                        }

                        cmdId = calculateIdCommand(dis.itemID);
                        menuItem = Items[cmdId - 1];

                        // Create new event arguments from known variables
                        System.Windows.Forms.DrawItemEventArgs ea = new DrawItemEventArgs(gr, SystemInformation.MenuFont, dis.rcItem, cmdId, state, foreColor, backColor);

                        if (menuItem.OnDrawItem != null)
                            menuItem.DoDrawItem(ea);
                        else
                            DoDrawItem(menuItem, ea);

                        // Marshal structure back to pointer if anything changes inside the structure. This is most unlikely 
                        // Marshal.StructureToPtr(dis, lParam, false);
                    }
                    finally
                    {
                        // Graphics.ReleaseHdc() is meant to be called after Graphics.GetHdc(). 
                        // If you use Graphics.FromHdc(), then Dispose() is all you need. 
                        gr.Dispose();
                    }
                    break;
                case (int)WM.WM_MEASUREITEM:
                    // For menu's the wParam equals zero
                    if (wParam != IntPtr.Zero)
                        return 0;

                    // We need to marshal the structure because we need access to the itemID in order to determine the menuitem
                    // this event is referring to
                    MEASUREITEMSTRUCT mis = (MEASUREITEMSTRUCT)Marshal.PtrToStructure(lParam, typeof(MEASUREITEMSTRUCT));

                    cmdId = calculateIdCommand(mis.itemID);
                    menuItem = Items[cmdId - 1];

                    if (menuItem.OnMeasureItem != null)
                    {
                        // Get access to the desktop window, this will be used to get the standard metrics
                        gr = Graphics.FromHwnd(IntPtr.Zero);
                        try
                        {
                            // Compensate for checkmarks in the menuitem
                            int defHeight = Math.Max(mis.itemHeight, SystemInformation.MenuCheckSize.Height + 3);

                            System.Windows.Forms.MeasureItemEventArgs ea = new MeasureItemEventArgs(gr, cmdId, defHeight);
                            ea.ItemWidth = mis.itemWidth;
                            menuItem.DoMeasureItem(ea);

                            // Only perform the marshalling if something has changed
                            if ((ea.ItemWidth != mis.itemWidth) || (ea.ItemHeight != mis.itemHeight))
                            {
                                mis.itemHeight = ea.ItemHeight;
                                mis.itemWidth = ea.ItemWidth;

                                // Marshal structure back to pointer for shell to use
                                Marshal.StructureToPtr(mis, lParam, false);
                            } // if
                        } // try
                        finally
                        {
                            gr.Dispose();
                        }
                    }

                    break;
            }
            return 0;
        }
        #endregion

        #region Virtuals
        /*
        protected virtual void DoMeasureItem(System.Windows.Forms.MeasureItemEventArgs ea)
        {
            SizeF size = ea.Graphics.MeasureString(m_items[ea.Index - 1].Text, SystemInformation.MenuFont);

            ea.ItemWidth = (int)size.Width;
            // Use image height if appropriate and adjust offset by 3 so item aligns with others
            ea.ItemHeight = Math.Max((int)size.Height, ea.ItemHeight);

        } // DoMeasureItem()
        */

        /// <summary>
        /// Performs the actual painting of the menu items
        /// </summary>
        /// <param name="menuItem">The menu item.</param>
        /// <param name="ea">The <see cref="System.Windows.Forms.DrawItemEventArgs"/> instance containing the event data.</param>
        protected virtual void DoDrawItem(ContextMenuItem menuItem, System.Windows.Forms.DrawItemEventArgs ea)
        {
            int x = ea.Bounds.Left;
            int y = ea.Bounds.Top;

            // Font verdana = new Font("Verdana", 20);

            // Perform default highlight and normal background painting
            ea.DrawBackground();

            // Add the standard image width to the horizontal offset so that menuitem is aligned with others
            x += System.Windows.Forms.SystemInformation.MenuCheckSize.Width + 1;
            // Adjust y offset by 1
            y += 1;

            ea.Graphics.DrawString(Items[ea.Index - 1].Text, ea.Font, new SolidBrush(ea.ForeColor), x, y);
        } // DoDrawItem(cmdId, eventArgs)

        /// <summary>
        /// Accepts the context of the popup for this extension
        /// </summary>
        /// <param name="filePath">The file path.</param>
        /// <returns>True in case the extension should be activated otherwise false</returns>
        /// <remarks>
        /// Descendants could override this method on order to provide custom logic
        /// </remarks>
        protected virtual bool AcceptContext(string filePath)
        {
            return true;
        }

        /// <summary>
        /// Populates the menu with menuitems
        /// </summary>
        /// <remarks>
        /// Descendants must override this method
        /// </remarks>
        protected virtual void InitializeMenu()
        {
            throw new Exception("The method or operation is not implemented.");
        }
        #endregion
    }
}
